package fc;

import java.awt.BorderLayout;
import java.awt.CheckboxMenuItem;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.image.RenderedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import javax.imageio.ImageIO;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import kjx.ui.UIHelper;
import fc.accessible.FCAccessible;
import fc.accessible.FCAccessibleContainer;
import fc.accessible.FCAccessibleObject;
import fc.accessible.FCAccessibleProperties;
import fc.awt.FCFractalPanel;
import fc.def.FCConstDef;

public class FCMain extends Frame implements FCConstDef {

	private static final String VERSION = "0.1";
    private static final String CONST_PROPERTIES_FILENAME = "FCFractal.properties";

    private static final FileFilter FC_EXTENSION_FILE_FILTER = new ExtensionFileFilter( "FC Parameter File", ".fc" );
    private static final FileFilter FCP_EXTENSION_FILE_FILTER = new ExtensionFileFilter( "FC Property File", ".properties" );
    private static final FileFilter PNG_EXTENSION_FILE_FILTER = new ExtensionFileFilter( "PNG Image File", ".png" );

    private Properties props;

    private HashMap accessible = new HashMap();
    private FCAccessibleContainer selected = new FCAccessibleContainer();

    private FCFractalPanel panel;

    private PopupMenu popup = new PopupMenu();
//    private CheckboxMenuItem  loadPaletteMi;

    private File lastLoadSaveFolder;
    
    public FCMain() {

        setTitle( "EffCee v" + VERSION );

        try {
            init();
        } catch(Exception e) {
            e.printStackTrace();
        }

        setVisible( true );

    }

    public static void main( String pArgs[] ) {
        new FCMain();
    }

    private Menu buildAccessiblePopupMenu( int pType, String pName, ArrayList pAccessible,
            int pDefault ) {

        Menu wMn = new Menu( pName );

        CbItemListener wIl = new CbItemListener( pType, pDefault );

        for( int a = 0; a < pAccessible.size(); a++ ) {

            CheckboxMenuItem wMi = new CheckboxMenuItem(
                ( (FCAccessible)pAccessible.get( a ) ).getDescription() );

            if ( a == pDefault ) {
                wMi.setState( true );
            }

            wMi.addItemListener( wIl );
            wMn.add( wMi );

        }

        popup.add( wMn );
        
        return wMn;

    }

    private void buildPaletteAdditionMenuItems( Menu pPaletteMenu ) {
    	
    	pPaletteMenu.addSeparator();

    	pPaletteMenu.add( UIHelper.createCheckboxMenuItem( "Load...", "LoadPalette", this ) ); 
    	
    }
    
    private void buildStaticPopupMenu() {
    	
        popup.addSeparator();

        popup.add( UIHelper.createCheckboxMenuItem( "Hide Parameters", this ) );
        popup.add( UIHelper.createMenuItem( "Reset", this ) );
        
        popup.addSeparator();
        
        Menu wMf = new Menu( "File" );
        
        wMf.add( UIHelper.createMenuItem( "Load Parameters", this ) );
        wMf.add( UIHelper.createMenuItem( "Save Parameters", this ) );
        
        popup.add( wMf );
        
        Menu wMh = new Menu( "Help" );
        
        wMh.add( UIHelper.createMenuItem( "How to...", this ) );
        wMh.add( UIHelper.createMenuItem( "About...", this ) );
        
        popup.add( wMh );
        
        popup.addSeparator();

        popup.add( UIHelper.createMenuItem( "Exit", this ) );
    	
    }

    private void checkboxMenuItemHideParameters( ItemEvent pEvent ) {
    	panel.setHideAccessibles( pEvent.getStateChange() == ItemEvent.SELECTED );
    }
    
    private void checkboxMenuItemLoadPalette( ItemEvent pEvent ) {
    	
    	CheckboxMenuItem wLpcbmi = (CheckboxMenuItem)pEvent.getSource();
    	
    	if ( loadPalette() ) {
    	
	    	// -- Clear existing checked palettes.
	    	Menu wPm = (Menu)wLpcbmi.getParent();
	    	
	    	// -- Only check everything but the last two items.
	    	for( int a = 0; a < wPm.getItemCount() - 2; a++ ) {
	    		( (CheckboxMenuItem)wPm.getItem( a ) ).setState( false );
	    	}
	    	
	    	wLpcbmi.setState( true );
	    	
    	} else {
    	
    		wLpcbmi.setState( false );

		}
    	
    }
    
    private void init() throws Exception {

        addWindowListener(
            new java.awt.event.WindowAdapter() {

                public void windowClosing(WindowEvent e) {
                    this_windowClosing(e);
                }
            }
        );

        panel = new FCFractalPanel( selected );

        if ( panel != null ) {
            add( panel, BorderLayout.CENTER );
            installPopup( panel );
        }

 //       panel.addActionListener( this );
        
        addKeyListener( panel );

        loadProperties();

        buildStaticPopupMenu();
        
        applySelection();

        setSize(
            getPropertyAsInt( PROP_FRAME_WIDTH, 512 ),
            getPropertyAsInt( PROP_FRAME_HEIGHT, 512 ) );

        
        
    }

    private void applySelection() {

/*        panel.setRenderer( (FCRenderer)selected.get( PROP_RENDERER ) );

        panel.getRenderer().setFractal( (FCFractal)selected.get( PROP_FRACTAL ) );

        FCPalette wPal = (FCPalette)selected.get( PROP_PALETTE );

        if ( panel.getRenderer().getPalette() != wPal ) {
            panel.getRenderer().setPalette( wPal );
            panel.getRenderer().getFractal().setIterations(
                (short)( wPal.getLength() - 1 ) );
        } */

        panel.repaint();

    }

    private int getPropertyAsInt( String pProp, int pDefault ) {

        try {
            return Integer.parseInt( props.getProperty( pProp ) );
        } catch( Exception pEx ) {
            return pDefault;
        }

    }
    
//    private FCFractal getSelectedFractal() {
//    	return (FCFractal)selected.getAccessible( FCAccessibleContainer.ACCESSIBLE_FRACTAL );
//    }
//
//    private FCRenderer getSelectedRenderer() {
//    	return (FCRenderer)selected.getAccessible( FCAccessibleContainer.ACCESSIBLE_RENDERER );
//    }

    private void installPopup( Component pComp ) {

        pComp.add( popup );

        pComp.addMouseListener(
            new MouseAdapter() {
                public void mousePressed( MouseEvent pEvent ) {
                    if ( ( pEvent.getModifiers() & MouseEvent.BUTTON3_MASK ) > 0 ) {
                        popup.show( pEvent.getComponent(), pEvent.getX(), pEvent.getY() );
                    }
                }
            }
        );

    }

    private void loadProperties() {

        try {

            props = new Properties();

            props.load( getClass().getClassLoader().getResourceAsStream( CONST_PROPERTIES_FILENAME ) );

            loadAccessibleType( props, FCAccessibleContainer.ACCESSIBLE_RENDERER );
            loadAccessibleType( props, FCAccessibleContainer.ACCESSIBLE_FRACTAL );
            
            buildPaletteAdditionMenuItems(
            	loadAccessibleType( props, FCAccessibleContainer.ACCESSIBLE_PALETTE, PROP_PALETTE_DEFAULT_CLASS )
            );

            
        } catch( Exception pEx ) {

 //           System.out.println( "Error loading properties: " + pEx.toString() );
            pEx.printStackTrace();

        }

    }

    private Menu loadAccessibleType( Properties pProps, int pType ) {
        return loadAccessibleType( pProps, pType, null );
    }

    private Menu loadAccessibleType( Properties pProps, int pType, String pDefaultClass ) {

        String wPre = FCAccessibleContainer.getTypeNameForTypeId( pType );

        ArrayList wAccessible = parseProperties( wPre, pProps, pDefaultClass );

        // -- Store accessibles
        accessible.put( wPre, wAccessible );

        // -- Add accessibles to context menu.
        return buildAccessiblePopupMenu(
            pType,
            pProps.getProperty( wPre + ".menu" ),
            wAccessible,
            getPropertyAsInt( wPre + ".default", 0 ) );

    }

    private void menuItemAbout( ActionEvent pEvent ) {
    	
    	JOptionPane.showMessageDialog( 
			FCMain.this, 
			"EffCee v" + VERSION + "\r\n\r\n" +
			"Created by: K. Fudalewski\r\n" +
			"Email: sirk_sytes@hotmail.com\r\n" +
			"Web: http://sirk.sytes.net\r\n",
			"About...",
			JOptionPane.PLAIN_MESSAGE );
    	
    }
    
    private void menuItemExit( ActionEvent pEvent ) {
    
    	FCMain.this.dispose();
        System.exit( 0 );
    	
    }
    
    private void menuItemHowto( ActionEvent pEvent ) {

    	JOptionPane.showMessageDialog( 
			FCMain.this, 
			"- Click and drag the fractal to move and explore.\r\n" +
			"- Click any parameter to edit.\r\n" +
			"- Double-click, or hold the ALT key and drag, to zoom in.\r\n" +
			"- Hold the CTRL key and click to zoom out.\r\n" +
			"- Press ESC to stop rendering at anytime.\r\n" +
			"- Press F5 to repaint fractal.\r\n",
			"How to...",
			JOptionPane.PLAIN_MESSAGE );

    }
    
    private void menuItemLoadParameters( ActionEvent pEvent ) {
    	loadParameters();
    }
    
    private void menuItemReset( ActionEvent pEvent ) {
    	
    	int wRc = JOptionPane.showConfirmDialog( 
   			FCMain.this,
   			"Reset fractal to default values?",
   			"Reset",
   			JOptionPane.OK_CANCEL_OPTION );
    	
    	if ( wRc == JOptionPane.OK_OPTION ) {
    		panel.getAccessibleContainer().getAccessible( FCAccessibleContainer.ACCESSIBLE_FRACTAL ).reset();
    		panel.repaint();
    	}
    	
    }
    
    private void menuItemSaveParameters( ActionEvent pEvent ) {
    	saveParameters();
    }
    
    private ArrayList parseProperties( String pPre, Properties pProp ) {
        return parseProperties( pPre, pProp, null );
    }

    private ArrayList parseProperties( String pPre, Properties pProp, String pDefClass ) {

        int wCount = 0;
        String wPre = pPre + '.';
        String wProp;

        String wPath = pProp.getProperty( wPre + "path" );

        ArrayList wInstances = new ArrayList();

        while ( ( wProp = props.getProperty( wPre + wCount ) ) != null ) {

            try {
                wInstances.add( FCAccessible.loadAccessible( wPath, wProp, pDefClass ) );
            } catch( Exception pEx ) {
//                System.out.println( "Error loading properties for: " + wProp + " ( " + pEx + " )" );
                pEx.printStackTrace();
            }

            wCount++;

        }

        return wInstances;

    }

/*    private ArrayList loadPalettes() {

        String wPath = props.getProperty( FCAccessible.ACCESSIBLE_PALETTE + ".path" );

        if ( wPath == null ) {
            System.out.println( "No palettes available!" );
            return null;
        }

        int wCount = 0;
        String wProp;

        ArrayList wPalettes = new ArrayList();

        while ( ( wProp = props.getProperty( PROP_PALETTE_DEF_PREFIX + wCount ) ) != null ) {

            try {

                Properties wProps = new Properties();
                wProps.load( new FileInputStream( new File( wPath, wProp + ".palette" ) ) );

                wPalettes.add( new FCPalette( wProps ) );

            } catch( Exception pEx ) {
                System.out.println( "Unable to load palette: " + wProp + " ( " + pEx + " )" );
            }

            wCount++;

        }

        return wPalettes;

    } */

    public void repaintFractal() {
        panel.repaint();
    }

    protected boolean loadPalette() {

    	JFileChooser wJf = new JFileChooser( lastLoadSaveFolder );
    	
    	wJf.setFileFilter( FCP_EXTENSION_FILE_FILTER );
    	
    	wJf.showOpenDialog( this );
    	
    	File wSf = wJf.getSelectedFile();
    	
    	if ( wSf != null ) {
    		
    		lastLoadSaveFolder = wSf.getParentFile();
    		
    		try {
    			
    			loadPalette( wSf );
    			
    			return true;
    			
    		} catch( IOException pEx ) {
    			JOptionPane.showMessageDialog( 
    				this, 
    				"An error occurred while loading: \r\n" + pEx.getLocalizedMessage(), 
    				"Error!",
    				JOptionPane.ERROR_MESSAGE );
    		}
    		
    	}
    	
    	return false;
    	
    }
    
    protected void loadPalette( File pPaletteFile ) throws IOException {

    	FCAccessibleProperties wAp = new FCAccessibleProperties();
    	
    	wAp.load( new FileInputStream( pPaletteFile ) );
    	
    	FCAccessible wAccessiblePalette = FCAccessible.loadAccessible( wAp, PROP_PALETTE_DEFAULT_CLASS );
    	
    	panel.getAccessibleContainer().setAccessible( FCAccessibleContainer.ACCESSIBLE_PALETTE, wAccessiblePalette );

    	panel.repaint();
    	
    }
    
    protected void loadParameters() {

    	JFileChooser wJf = new JFileChooser( lastLoadSaveFolder );
    	
    	wJf.setFileFilter( FC_EXTENSION_FILE_FILTER );
    	
    	wJf.showOpenDialog( this );
    	
    	File wSf = wJf.getSelectedFile();
    	
    	if ( wSf != null ) {
    		
    		lastLoadSaveFolder = wSf.getParentFile();
    		
    		try {
    			loadParameters( wSf );
    		} catch( IOException pEx ) {
    			JOptionPane.showMessageDialog( 
    				this, 
    				"An error occurred while loading: \r\n" + pEx.getLocalizedMessage(), 
    				"Error!",
    				JOptionPane.ERROR_MESSAGE );
    		}
    	}
    	
    }
    
    protected void loadParameters( File pFile ) throws IOException {
    	
    	boolean wIfp = panel.isSettingIterationsFromPalette();
    	
    	panel.setIterationsFromPalette( false );
    	
    	try {
    	
	    	Properties wProps = new Properties();
	    	
	    	wProps.load( new FileInputStream( pFile ) );
	    	
	    	for( int a = 0; a < FCAccessibleContainer.ACCESSIBLE_TYPE_COUNT; a++ ) {
	    		
	    		String wAtn = FCAccessibleContainer.getTypeNameForTypeId( a );
	    		String wTid = wProps.getProperty( wAtn );
	
	    		String wPrefix = wAtn + "." + wTid + ".";
	    		
	    		List wAtl = (List)accessible.get( wAtn );
	    		
	    		for( int b = 0; b < wAtl.size(); b++ ) {
	    			
	    			FCAccessible wAccessible = (FCAccessible)wAtl.get( b );
	    			
	    			if ( wAccessible.getId().equals( wTid ) ) {
	    				
		    			List wAos = (List)wAccessible.getAccessibleObjects();
		    			
		    			for( int c = 0; c < wAos.size(); c++ ) {
		    				
		    				FCAccessibleObject wAo = (FCAccessibleObject)wAos.get( c );
		    				
		    				String wValue = wProps.getProperty( wPrefix + wAo.getId() );
		    				
		    				if ( wValue != null ) {
		    					wAo.setValue( wAo.stringToType( wValue ) );
		    				}
		    				
		    			}
		    			
		    			selected.setAccessible( a, wAccessible );
		    			
		    			break;
		    			
	    			}
	    			
	    		}
		    	
	    	}
	
	    	panel.repaint();
	    
    	} finally {
    		
    		panel.setIterationsFromPalette( wIfp );
    		
    	}
    	
    }
    
    protected void saveParameters() {
    	
    	JFileChooser wJf = new JFileChooser( lastLoadSaveFolder );
    	
    	wJf.setFileFilter( FC_EXTENSION_FILE_FILTER );
    	
    	wJf.showSaveDialog( this );
    	
    	File wSf = wJf.getSelectedFile();
    	
    	if ( wSf != null ) {
    		
    		// -- Append '.fc' extension.
    		if ( !wSf.getName().toLowerCase().endsWith( ".fc" ) ) {
    			wSf = new File( wSf.getParentFile(), wSf.getName() + ".fc" );
    		}
    		
    		lastLoadSaveFolder = wSf.getParentFile();
    		
    		try {
    			saveParameters( wSf );
    		} catch( IOException pEx ) {
    			JOptionPane.showMessageDialog( 
    				this, 
    				"An error occurred while saving: \r\n" + pEx.getLocalizedMessage(), 
    				"Error!",
    				JOptionPane.ERROR_MESSAGE );
    		}
    	}
    	
    }
    
    protected void saveParameters( File pFile ) throws IOException {
    	
    	Properties wProps = new Properties();
    	FCAccessible[] wAccessibles = selected.getAccessibles();
    	
    	for( FCAccessible wAccessible : wAccessibles ) {
    	
    		String wType = FCAccessibleContainer.getTypeNameForTypeId( wAccessible.getType() );
    		String wId   = wAccessible.getId();
    		
	    	// -- Put the type with id.
	    	wProps.put( wType, wId );

	    	String wPrefix = wType + "." + wId + ".";
	    	
	    	for( int a = 0; a < wAccessible.getAccessibleObjects().size(); a++ ) {
	    		
	    		FCAccessibleObject wAo = (FCAccessibleObject)wAccessible.getAccessibleObjects().get( a );
	    		
	    		wProps.put( wPrefix + wAo.getId(), String.valueOf( wAo.getValue() ) );
	    		
	    	}
	    	
    	}
    	
    	wProps.store( new FileOutputStream( pFile ), null );
    	
    }

    protected void saveToPng() {
    	
    	JFileChooser wJf = new JFileChooser( lastLoadSaveFolder );
    	
    	wJf.setFileFilter( PNG_EXTENSION_FILE_FILTER );
    	
    	wJf.showSaveDialog( this );
    	
    	File wSf = wJf.getSelectedFile();
    	
    	if ( wSf != null ) {
    		
    		lastLoadSaveFolder = wSf.getParentFile();
    		
    		try {
    			saveToPng( wSf );
    		} catch( IOException pEx ) {
    			JOptionPane.showMessageDialog( 
    				this, 
    				"An error occurred while saving: \r\n" + pEx.getLocalizedMessage(), 
    				"Error!",
    				JOptionPane.ERROR_MESSAGE );
    		}
    	}
    	
    }
    
    protected void saveToPng( final File pFile ) throws IOException {

    	new Thread(
    		new Runnable() {
    			public void run() {
    		    	
    		    	RenderedImage wRi = panel.createSnapshot( 2048,	2048 );
    		    	
    		    	try {
    		    		ImageIO.write( wRi, "png", pFile );
    		    	} catch( IOException pEx ) {
    		    		pEx.printStackTrace();
    		    	}

    			}
    		}
    	).start();
    	
    	WaitDialog wWd = new WaitDialog( 
    		this, 
    		"Rendering...", 
    		"Please wait while rendering fractal...",
    		new ActionListener() {
				public void actionPerformed( ActionEvent pEvent ) {
					panel.getRenderer().stop();
				}
    		}
    	);
    	
    	wWd.setVisible( true );
    	
    }
    
    
    void this_windowClosing(WindowEvent e) {
        dispose();
        System.exit( 0 );
    }

//    public void actionPerformed( ActionEvent pEvent ) {

//    	System.out.println( "AP: " + pEvent );
    	
//        FCRenderer wRenderer = getSelectedRenderer();
//
//        if ( wRenderer != null ) {
//            wRenderer.refreshAccessible();
//        }

//        rendererDialog.updateFields();

//        FCFractal wFractal = getSelectedFractal();
//
//        if ( wFractal != null ) {
//            wFractal.refreshAccessible();
//        }

//        fractalDialog.updateFields(); 

 //   }
    
    public void update( Graphics pGrp ) {
    	paint( pGrp );
    }

    private class CbItemListener implements ItemListener {

        private int type;

        public CbItemListener( int pType, int pDefault ) {
            type = pType;
            setSelected( pDefault );
        }

        public void itemStateChanged( ItemEvent pEvent ) {

            CheckboxMenuItem wCbmi = (CheckboxMenuItem)pEvent.getSource();

            Menu wMc = (Menu)wCbmi.getParent();

            int wIdx = 0;
            int wCnt = wMc.getItemCount();
            
            if ( type == FCAccessibleContainer.ACCESSIBLE_PALETTE ) {
            	wCnt -= 2;
            }
            
            for( int a = 0; a < wCnt; a++ ) {

                CheckboxMenuItem wCbmi1 = (CheckboxMenuItem)wMc.getItem( a );

                if ( wCbmi != wCbmi1 ) {
                    wCbmi1.setState( false );
                } else {
                    wIdx = a;
                }

            }

           	setSelected( wIdx );
           	
            if ( type == FCAccessibleContainer.ACCESSIBLE_PALETTE ) {
               	// -- Clicked an internal palette, then clear checked state on 'Load...' item.
               	( (CheckboxMenuItem)wMc.getItem( wMc.getItemCount() - 1 ) ).setState( false );
            }

        }

        public int getType() {
            return type;
        }

        private void setSelected( int pIdx ) {

            String wPre = FCAccessibleContainer.getTypeNameForTypeId( type );

            List wAl = (List)accessible.get( wPre );
            
            if ( pIdx < wAl.size() ) {

            	selected.setAccessible( type, (FCAccessible)wAl.get( pIdx ) );
            
            	applySelection();
            	
            }

        }

    }
    
    private static class ExtensionFileFilter extends FileFilter {

    	private String description;
    	private String extension;
    	
    	public ExtensionFileFilter( String pDesc, String pExtension ) {
    		description = pDesc;
    		extension = pExtension.toLowerCase();
    	}
    	
		public boolean accept( File pPathname ) {
			return pPathname.isDirectory() || pPathname.getName().toLowerCase().endsWith( extension );
		}

		@Override
		public String getDescription() {
			// TODO Auto-generated method stub
			return description + " (*" + extension + ")";
		}
    	
    }

    private static class WaitDialog extends JDialog {
    	
    	public WaitDialog( Frame pParent, String pTitle, String pMessage, ActionListener pCancel ) {
    		
    		super( pParent, true );
    		
    		setTitle( pTitle );
    		
    		init( pMessage, pCancel );
    		
    	}
    	
    	private void init( String pMessage, ActionListener pCancel ) {
    		
    		setLayout( new BorderLayout() );
    		
    		add( new JLabel( pMessage ) );
    		
    		JButton wCb = new JButton( "Canel" );
    		
    		wCb.addActionListener( pCancel );
    		
    		add( wCb, BorderLayout.SOUTH );
    		
    	}
    	
    }
    
}


