package fc.accessible;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

public abstract class FCAccessible {

    private static final String PROP_PRE_ACCESSIBLE = "accessible.";

    private FCAccessibleContainer container;

    private List<FCAccessibleObject> accessibleObjects;

    private String id;
    private String desc;

    public FCAccessible( Properties pProp ) {

    	id = pProp.getProperty( "id" );

    	desc = parseDescription( pProp );
    	
    }

    public void apply() {
    	
    	for ( int a = 0; a < accessibleObjects.size(); a++ ) {
    	  ( (FCAccessibleObject)accessibleObjects.get( a ) ).apply();
    	}
    	
    }

//    protected void apply( ArrayList pAf ) {
//
//  for ( int a = 0; a < pAf.size(); a++ ) {
//  ( (FCAccessibleObject)pAf.get( a ) ).apply();
//}
//
//    }

    public FCAccessibleContainer getAccessibleContainer() {
        return container;
    }

    public List<FCAccessibleObject> getAccessibleObjects() {
        return accessibleObjects;
    }

    public String getDescription() {
        return desc;
    }

    public String getId() {
    	return id;
    }
    
    public abstract int getType();
    
    public void clearModified() {

        for ( int a = 0; a < accessibleObjects.size(); a++ ) {
            ( (FCAccessibleObject)accessibleObjects.get( a ) ).setModified( false );
        }

    }

    public boolean isModified() {

        for ( int a = 0; a < accessibleObjects.size(); a++ ) {
            if ( ( (FCAccessibleObject)accessibleObjects.get( a ) ).isModified() ) {
                return true;
            }
        }

        return false;

    }

    protected String parseDescription( Properties pProp ) {
    	return pProp.getProperty( "description" );
    }
    
    private void parseAccessibleProperties( FCAccessibleProperties pProp ) {

        accessibleObjects = new ArrayList<FCAccessibleObject>();

        parseAccessibleProperties( pProp, pProp, accessibleObjects );
        
    }
    
    private void parseAccessibleProperties( Properties pRoot, FCAccessibleProperties pProp, List<FCAccessibleObject> pAos ) {

    	// -- Check for default properties, and process those first.
    	if ( pProp.getDefaults() != null ) {
    		parseAccessibleProperties( pRoot, pProp.getDefaults(), pAos );
    	}
    	
    	String wKey;
        int wCnt = 0;
        
        while ( ( wKey = pProp.getProperty( PROP_PRE_ACCESSIBLE + wCnt, false ) ) != null ) {

            // -- Create accessible object for field
            try {
            	accessibleObjects.add( new FCAccessibleObject( this, wKey, pRoot ) );
            } catch( Exception pEx ) {
                throw new FCAccessibleException( pEx );
            }

            wCnt++;

        }
	        
    }

    public void refresh() {

        for ( int a = 0; a < accessibleObjects.size(); a++ ) {
            ( (FCAccessibleObject)accessibleObjects.get( a ) ).refresh();
        }

    }

    public void reset() {
    	
    	for ( int a = 0; a < accessibleObjects.size(); a++ ) {
    	    ( (FCAccessibleObject)accessibleObjects.get( a ) ).clear();
    	}
    	
    	apply();
    	
    }
    
    void setAccessibleContainer( FCAccessibleContainer pContainer ) {
        container = pContainer;
    }

    public static FCAccessible loadAccessible( String pPath, String pName )
            throws Exception {

        return loadAccessible( pPath, pName, null );

    }

    public static FCAccessible loadAccessible( String pPath, String pId, String pDefClass ) {

    	return loadAccessible( loadAccessibleProperties( pPath, pId ), pDefClass );
    	
    }
    
    public static FCAccessible loadAccessible( FCAccessibleProperties pAp, String pDefClass ) {

        String wClassName = pAp.getProperty( "class" );

        if ( wClassName == null ) {
            wClassName = pDefClass;
        }

        try {
        
	        Class wClass = Class.forName( wClassName );
	
	        Constructor wCon = wClass.getConstructor( new Class[] { Properties.class } );
	
	        FCAccessible wAccessible = (FCAccessible)wCon.newInstance( new Object[] { pAp } );
	        
	        wAccessible.parseAccessibleProperties( pAp );
	        
	        return wAccessible;
	        
        } catch( Exception pEx ) {
        	throw new FCAccessibleException( "Error loading accessible.", pEx );
        }

    }

    private static FCAccessibleProperties loadAccessibleProperties( String pPath, String pId ) {

        FCAccessibleProperties wProp = new FCAccessibleProperties();

        loadAccessibleProperties( pPath, pId, wProp ); 

        wProp.put( "id", pId );
        
        return wProp;

    }

    private static void loadAccessibleProperties( String pPath, String pId, FCAccessibleProperties pProp ) {

    	String wPfp = pPath + '/' + pId + ".properties";
    	InputStream wIs = ClassLoader.getSystemClassLoader().getResourceAsStream( wPfp );
    	
    	if ( wIs == null ) {
    		throw new FCAccessibleException( "Properties not found: " + wPfp );
    	}
    
    	try {
    		pProp.load( wIs );
    	} catch( IOException pEx ) {
    		throw new FCAccessibleException( "Error loading accesible properties: '" + wPfp + "'", pEx );
    	}

        String wExt;

        if ( ( wExt = pProp.getProperty( "extends" ) ) != null && ( pId == null || !wExt.equals( pId ) ) ) {

        	pProp.setDefaults( new FCAccessibleProperties() );
        	
            loadAccessibleProperties( pPath, wExt, pProp.getDefaults() );

        }

    }

}
