package fc.accessible;

import java.util.ArrayList;

public class FCAccessibleContainer {

    public static final int ACCESSIBLE_TYPE_COUNT = 3;

    public static final int ACCESSIBLE_RENDERER = 0;
    public static final int ACCESSIBLE_FRACTAL  = 1;
    public static final int ACCESSIBLE_PALETTE  = 2;

    private FCAccessible[] accessibles = new FCAccessible[ ACCESSIBLE_TYPE_COUNT ];
    private ArrayList[]    listeners   = new ArrayList[ ACCESSIBLE_TYPE_COUNT ];

    private String language = "en";

    public void apply() {
    	
    	for( int a = 0; a < accessibles.length; a++ ) {
    		accessibles[ a ].apply();
    	}
    	
    }
    
    public FCAccessible getAccessible( int pAccessibleType ) {
        return (FCAccessible)accessibles[ pAccessibleType ];
    }

    public FCAccessible[] getAccessibles() {
    	return accessibles;
    }
    
    public String getLanguage() {
        return language;
    }

    public void setAccessible( int pAt, FCAccessible pAccessible ) {

        accessibles[ pAt ] = pAccessible;

        pAccessible.setAccessibleContainer( this );

        if ( listeners[ pAt ] != null ) {

            ArrayList wAll = listeners[ pAt ];

            for( int a = 0; a < wAll.size(); a++ ) {
                ( (FCAccessibleChangeListener)wAll.get( a ) ).accessibleChanged(
                    this, pAt );
            }

        }

    }

    public void addAccessibleChangeListener( int pAt, FCAccessibleChangeListener pAcl ) {

        if ( listeners[ pAt ] == null ) {
            listeners[ pAt ] = new ArrayList();
        }

        listeners[ pAt ].add( pAcl );

    }

    
    public void refresh() {
    	
    	for( int a = 0; a < accessibles.length; a++ ) {
    		accessibles[ a ].refresh();
    	}
    	
    }
    
    public void removeAccessibleChangeListener( int pAt, FCAccessibleChangeListener pAcl ) {

        if ( listeners[ pAt ] != null ) {
            listeners[ pAt ].remove( pAcl );
        }

    }

    public static String getTypeNameForTypeId( int pAt ) {

        switch( pAt ) {

            case ACCESSIBLE_RENDERER:
                return "renderer";
            case ACCESSIBLE_FRACTAL:
                return "fractal";
            case ACCESSIBLE_PALETTE:
                return "palette";

        }

        return null;

    }

}