package fc.awt;

import java.awt.Color;
import java.awt.Container;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.DefaultButtonModel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import fc.accessible.FCAccessibleObject;

public class FCAccessibleEditor extends Container {

    private static final Font  INFO_FONT      =
        new Font( "dialog", Font.PLAIN, 11 );
    private static final Color INFO_COLOR = Color.green;
    
    private JTextField accessibleEditor;
    private JButton    accessibleEditorIncrease;
    private JButton    accessibleEditorDecrease;
    private JTextField accessibleEditorStep;
    private JButton    accessibleEditorDefault;

    private boolean right = false;
    
    private FCAccessibleObject ao;
    
    private List<EditingListener> editingListeners = new ArrayList<EditingListener>();
    
    public FCAccessibleEditor() {
    	
    	super();
    	
    	init();
    	
    }
    
    public void addEditingListener( EditingListener pListener ) {
    	editingListeners.add( pListener );
    }
    
    private void applyAccessibleEditorScheme( JComponent pComponent ) {
    	
    	pComponent.setOpaque( true );
    	pComponent.setBorder( BorderFactory.createLineBorder( INFO_COLOR ) );
					
    	pComponent.setBackground( Color.darkGray );
    	pComponent.setForeground( INFO_COLOR );
    	pComponent.setFont( INFO_FONT );
    	
//    	pComponent.setVisible( false );
    	
    }
    
    public void doLayout() {
    	
    	accessibleEditor.setSize( 128, getHeight() );
    	accessibleEditorDecrease.setSize( 16, getHeight() );
    	accessibleEditorIncrease.setSize( 16, getHeight() );
    	accessibleEditorStep.setSize( 48, getHeight() );
    	accessibleEditorDefault.setSize( 16, getHeight() );
    	
    	if ( right ) {
    		
//    		accessibleEditor.setHorizontalAlignment( JTextField.LEFT );
//    		accessibleEditorStep.setHorizontalAlignment( JTextField.LEFT );
    		
    		accessibleEditorDefault.setLocation( 0, 0 );
    		accessibleEditorStep.setLocation( 17, 0 );
    		accessibleEditorDecrease.setLocation( 66, 0 );
    		accessibleEditorIncrease.setLocation( 83, 0 );
    		accessibleEditor.setLocation( 100, 0 );
    		
    	} else {
    		
//    		accessibleEditor.setHorizontalAlignment( JTextField.RIGHT );
//    		accessibleEditorStep.setHorizontalAlignment( JTextField.RIGHT );
    		
    		accessibleEditor.setLocation( 0, 0 );
    		accessibleEditorIncrease.setLocation( 129, 0 );
    		accessibleEditorDecrease.setLocation( 146, 0 );
    		accessibleEditorStep.setLocation( 163, 0 );
    		accessibleEditorDefault.setLocation( 212, 0 );
    		
    	}
    	
    }
    
    protected void commitEditing() {
    	
    	for( EditingListener wEl : editingListeners ) {
    		wEl.editingCommitted( this );
    	}
	    	
    }
    
    public FCAccessibleObject getAccessibleObject() {
    	return ao;
    }
    
//    public Object getValue() {
//    	return value;
//    }
    
    private void init() {
    	
    	setLayout( null );
    	
        accessibleEditor = new JTextField();

        applyAccessibleEditorScheme( accessibleEditor );
        
        accessibleEditor.setCaretColor( INFO_COLOR );
        
        accessibleEditor.addKeyListener( 
        	new KeyAdapter() {
        	    public void keyPressed( KeyEvent pEvent ) {
        	    	//System.out.println( "AEKE: " + pEvent );
        	    	if ( pEvent.getKeyCode() == KeyEvent.VK_ESCAPE ) {
        	    		stopEditing( false );
        	    	} else if ( pEvent.getKeyCode() == KeyEvent.VK_ENTER ) {
        	    		stopEditing( true );
        	    	} 
        	    }
        	}
        );
        
        accessibleEditorStep = new JTextField();
        
        applyAccessibleEditorScheme( accessibleEditorStep );
        
        accessibleEditorStep.setCaretColor( INFO_COLOR );
        
        accessibleEditorStep.getDocument().addDocumentListener(
        	new DocumentListener() {

				public void changedUpdate( DocumentEvent pEvent ) {
					applyStepChange();
				}

				public void insertUpdate(DocumentEvent e) {
					applyStepChange();
				}

				public void removeUpdate(DocumentEvent e) {
					applyStepChange();
				}

				private void applyStepChange() {
					
					try {
						ao.setStep( ao.stringToType( accessibleEditorStep.getText() ) );
						accessibleEditorStep.setForeground( Color.green );
					} catch( Exception pEx ) {
						accessibleEditorStep.setForeground( Color.red );
					}
					
				}
				
        	}
        	
    	);
        
        accessibleEditorIncrease = new JButton();
        accessibleEditorIncrease.setModel( new AutoFireButtonModel() );
        
        applyAccessibleEditorScheme( accessibleEditorIncrease );
        
        accessibleEditorIncrease.setFocusPainted( false );
        accessibleEditorIncrease.setRequestFocusEnabled( false );
        
        accessibleEditorIncrease.addActionListener(
        	new ActionListener() {
				public void actionPerformed( ActionEvent pEvent ) {
					ao.increase();
					updateFields();
					commitEditing();
				}
        	}
        );
        
        accessibleEditorDecrease = new JButton();
        accessibleEditorDecrease.setModel( new AutoFireButtonModel() );
        accessibleEditorDecrease.setRequestFocusEnabled( false );

        applyAccessibleEditorScheme( accessibleEditorDecrease );
        
        accessibleEditorDecrease.setFocusPainted( false );
        
        accessibleEditorDecrease.addActionListener(
        	new ActionListener() {
				public void actionPerformed( ActionEvent pEvent ) {
					ao.decrease();
					updateFields();
					commitEditing();
				}
        	}
        );
        
        accessibleEditorDefault = new JButton( "d" );
        accessibleEditorDefault.setModel( new AutoFireButtonModel() );
        
        applyAccessibleEditorScheme( accessibleEditorDefault );
        
        accessibleEditorDefault.setFocusPainted( false );
        accessibleEditorDefault.setRequestFocusEnabled( false );
        
        accessibleEditorDefault.addActionListener(
        	new ActionListener() {
				public void actionPerformed( ActionEvent pEvent ) {
					ao.clear();
					updateFields();
					commitEditing();
				}
        	}
        );
        
		accessibleEditor.setHorizontalAlignment( JTextField.RIGHT );
		accessibleEditorStep.setHorizontalAlignment( JTextField.RIGHT );
        
        add( accessibleEditor );
        add( accessibleEditorStep );
        add( accessibleEditorIncrease );
        add( accessibleEditorDecrease );
        add( accessibleEditorDefault );
    	
    }

    public boolean isRightAligned() {
    	return right;
    }
    
    public void removeEditingListener( EditingListener pListener ) {
    	editingListeners.remove( pListener );
    }
    
    public void requestFocus() {
    	accessibleEditor.requestFocus( true );
    }
    
    public void selectAll() {
    	accessibleEditor.selectAll();
    }
    
    public void setEditingListener( EditingListener pListener ) {
    	editingListeners.clear();
    	addEditingListener( pListener );
    }
    
	public void setRightAligned( boolean pRight ) {
		right = pRight;
	}
    
    public void setSize( int pWidth, int pHeight ) {
		
		super.setSize( pWidth, pHeight );
		
		doLayout();
		
	}
	
//	public void setStep( Object pStep ) {
//		
//		if ( pStep instanceof Number ) {
//			step = pStep;
//		} else {
//			throw new IllegalArgumentException( "Must be a number!" );
//		}
//		
//		accessibleEditorStep.setText( String.valueOf( pStep ) );
//		
//	}
	
	public void setAccessibleObject( FCAccessibleObject pAo ) {

		ao = pAo;
		
		updateFields();
		
	}

	public void stopEditing( boolean pCommit ) {
		
		if ( pCommit && updateAccessibleObject() ) {
			commitEditing();
		}
		
    	for( EditingListener wEl : editingListeners ) {
    		wEl.editingEnded( this );
    	}
		
	}
	
//	public void paint( Graphics pGrp ) {
//		// -- Do nothing.
//	}
	
	public void update( Graphics pGrp ) {
		// -- Do nothing.
	}
	
	protected boolean updateAccessibleObject() {
		
		if ( ao == null ) {
			return false;
		}
		
		String wVal = accessibleEditor.getText();
		
		if ( wVal != null && !wVal.equals( String.valueOf( ao.getValue() ) ) ) {
			
			try {
				
				ao.setValue( ao.stringToType( wVal ) );
				
				return true;
				
			} catch( Exception pEx ) {
				pEx.printStackTrace();
			}

		}
		
		return false;
		
	}
	
	protected void updateFields() {
		
		if ( ao != null ) {
			
			accessibleEditor.setText( String.valueOf( ao.getValue() ) );
			accessibleEditorStep.setText( String.valueOf( ao.getStep() ) );
			
			accessibleEditorDecrease.setText( ao.isStepMultiplied() ? "/" : "-" );
			accessibleEditorIncrease.setText( ao.isStepMultiplied() ? "*" : "+" );
			
			accessibleEditor.requestFocus( true );
			accessibleEditor.selectAll();
			
		} else {
			accessibleEditor.setText( "" );
			accessibleEditorStep.setText( "" );
		}
		
	}
	
	private static class AutoFireButtonModel extends DefaultButtonModel implements Runnable {
	
	    public void setPressed( boolean pState ) {
	    	
	    	super.setPressed( pState );
	    	
	    	synchronized( AutoFireButtonModel.this ) {
	    		
		    	if ( isPressed() ) {
			    	new Thread( this ).start();
		    	} else {
		    		AutoFireButtonModel.this.notify();
		    	}
		    	
	    	}
	    	
	    }

	    public void run() {
	    	
	    	synchronized( AutoFireButtonModel.this ) {
	    	
	    		try {
		    		wait( 500 );
		    	} catch( InterruptedException pEx ) {
		    		// -- Do nothing.
		    	}

		    	while( isPressed() ) {
		    		
		    		fireActionPerformed(
	                    new ActionEvent(
	                    	this, 
	                    	ActionEvent.ACTION_PERFORMED,
                            getActionCommand() 
                        )
                    );
		    		
		    		try {
		    			wait( 50 );
		    		} catch( InterruptedException pEx ) {
		    			// -- Do nothing.
		    		}
		    		
		    	}
	    	
	    	}
	    	
	    }
	    
	}
	
	public static interface EditingListener {
		
		public void editingCommitted( FCAccessibleEditor pEditor );
		
		public void editingEnded( FCAccessibleEditor pEditor );
		
	}
	
	
}
