/* ----------------------------------------------------------------------------
 * File:       fc/awt/FCFractalPanel.java
 *
 * Author:     Unknown
 *
 * Copyright (c) - 1999, Technologies Interactives Mediagrif Inc.
 *
 * History
 * ----------------------------------------------------------------------------
 *  Date         User               Description
 * ----------------------------------------------------------------------------
 * 2000/02/11   Kris Fudalewski     Creation.
 * ----------------------------------------------------------------------------
 */
package fc.awt;

import java.awt.AWTEvent;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import java.util.ArrayList;
import java.util.List;

import javax.swing.ToolTipManager;

import com.sun.xml.internal.bind.v2.runtime.reflect.Accessor.SetterOnlyReflection;

import fc.accessible.FCAccessible;
import fc.accessible.FCAccessibleChangeListener;
import fc.accessible.FCAccessibleContainer;
import fc.accessible.FCAccessibleObject;
import fc.fractal.FCFractal;
import fc.renderer.FCPalette;
import fc.renderer.FCRenderer;

public class FCFractalPanel extends Container implements ComponentListener,
        FCAccessibleChangeListener, KeyListener, MouseListener, MouseMotionListener {

    private static final Font  INFO_FONT      =
        new Font( "dialog", Font.PLAIN, 11 );
    private static final Font  INFO_FONT_BOLD =
        new Font( "dialog", Font.BOLD, 13 );
    private static final Color INFO_COLOR = Color.green;
    private static final Color INFO_HIGHLIGHT_COLOR = Color.orange;

    private static final int POSITION_TOP    = 1;
    private static final int POSITION_BOTTOM = 2;
    private static final int POSITION_LEFT   = 4;
    private static final int POSITION_RIGHT  = 8;
    
    private FCAccessibleContainer accessibleContainer = null;
    private AccessibleHandler[] accessibleHandlers = new AccessibleHandler[ FCAccessibleContainer.ACCESSIBLE_TYPE_COUNT ];

    private FCAccessibleEditor accessibleEditor;
    
    private boolean hideAccessibles = false;
    
    private Painter painter;
    private Updater updater;
    
    private boolean forceRepaint = false;

    private int lastSz = -1;
    private int dragX;
    private int dragY;

    // -- Blit scrolling variables
    private double fdox;
    private double fdoy;
    private int ddx;
    private int ddy;

    private boolean accessibleEvent = false;
    private boolean magnify = false;
    
    private boolean setIterationsFromPalette = true;
    
    public FCFractalPanel( FCAccessibleContainer pAc ) {
        
    	super();
        
        accessibleContainer = pAc;
        
        init();
        
    }

    public boolean getAccesiblesHidden() {
    	return hideAccessibles;
    }
    
    public void accessibleChanged( FCAccessibleContainer pAc, int pType ) {

    	if ( setIterationsFromPalette && pType == FCAccessibleContainer.ACCESSIBLE_PALETTE ) {
    	
	        FCFractal wFrac = (FCFractal)pAc.getAccessible( FCAccessibleContainer.ACCESSIBLE_FRACTAL );
	        FCPalette wPal = (FCPalette)pAc.getAccessible( FCAccessibleContainer.ACCESSIBLE_PALETTE );
	
//	        System.out.println( "AC: " + pAc + " " + pType + " PL: " + wPal.getLength() );
	
	        wFrac.setIterations( (short)( wPal.getLength() - 1 ) );
	        wFrac.refresh();
	        
    	}
    	
    	buildAccessibleHandlers( pAc, pType );
    	
    	repaint();
    	
    }

    private void buildAccessibleHandlers( FCAccessibleContainer pAc, int pType ) {
    	
    	FCAccessible wAc = pAc.getAccessible( pType );
    	
    	int wPos = 0;
    	
    	if ( pType == FCAccessibleContainer.ACCESSIBLE_FRACTAL ) {
    		wPos = POSITION_TOP | POSITION_LEFT;
    	} else if ( pType == FCAccessibleContainer.ACCESSIBLE_RENDERER ) {
    		wPos = POSITION_TOP | POSITION_RIGHT;
    	} else if ( pType == FCAccessibleContainer.ACCESSIBLE_PALETTE ) {
    		wPos = POSITION_BOTTOM | POSITION_LEFT;
    	}

    	if ( wPos != 0 ) {
    		accessibleHandlers[ pType ] = new AccessibleHandler( wAc, wPos );
    	}
    	
    }
    
//    public void addActionListener( ActionListener pAl ) {
//
//        if ( actionListeners == null ) {
//            actionListeners = new ArrayList();
//        }
//
//        actionListeners.add( pAl );
//
//    }
//
//    public void removeActionListener( ActionListener pAl ) {
//
//        if ( actionListeners != null ) {
//            actionListeners.remove( pAl );
//        }
//
//    }

    private boolean canRender() {
        return accessibleContainer != null;
    }

//    private void createBuffer() {
//        
//    	buffer = getGraphicsConfiguration().createCompatibleImage(
//            getSize().width, 
//            getSize().height );
//        
//    }
    
//    private void endAnyEditing( boolean pCommit ) {
//    	
//    	for( AccessibleHandler wAh : accessibleHandlers ) {
//    		if ( wAh != null ) {
//    			wAh.endEditing( pCommit );
//    		}
//    	}
//    	
//    }

    private boolean handleAccessibleEvent( MouseEvent pEvent ) {

    	if ( hideAccessibles ) {
    		return false;
    	}
    	
    	boolean eh = false;
    	
    	for( int a = 0; a < accessibleHandlers.length; a++ ) {
    		
    		if ( accessibleHandlers[ a ] != null ) {
    	        if ( accessibleHandlers[ a ].handleEvent( pEvent ) ) {
    	        	eh = true;
    	        }
    		}
    		
    	}
    	
    	return eh;

    }
    
    private void paintAccessibles( Graphics pGrp ) {

    	if ( hideAccessibles ) {
    		return;
    	}
    	
    	boolean sh = false;
    	
    	for( int a = 0; a < accessibleHandlers.length; a++ ) {
    		
    		if ( accessibleHandlers[ a ] != null ) {
    			if ( accessibleHandlers[ a ].paint( pGrp ) ) {
    				sh = true;
    			}
    		}
    		
    	}
    	
//    	System.out.println( "HC: " + sh );
    	if ( !isBusy() ) {
	    	if ( sh ) {
	    		setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
	    	} else {
	    		setCursor( Cursor.getDefaultCursor() ); // getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
	    	}
    	}

    	super.paint( pGrp );

    }

    private void init() {

        setLayout( null );
        setIgnoreRepaint( true );
        
        addComponentListener( this );
        addMouseListener( this );
        addMouseMotionListener( this );
        addKeyListener( this );

        getAccessibleContainer().addAccessibleChangeListener(
            FCAccessibleContainer.ACCESSIBLE_RENDERER, this );
        getAccessibleContainer().addAccessibleChangeListener(
            FCAccessibleContainer.ACCESSIBLE_FRACTAL, this );
        getAccessibleContainer().addAccessibleChangeListener(
            FCAccessibleContainer.ACCESSIBLE_PALETTE, this );

//        setCursor( new Cursor( Cursor.CROSSHAIR_CURSOR ) );

        initAccessibleEditor();
        initUpdateAndPaintingThreads();
        
    }

    private void initAccessibleEditor() {
    	
        accessibleEditor = new FCAccessibleEditor();

        accessibleEditor.setVisible( false );
        
//        applyAccessibleEditorScheme( accessibleEditor );
//        
//        accessibleEditor.setCaretColor( INFO_COLOR );
//        
//        accessibleEditor.addKeyListener( 
//        	new KeyAdapter() {
//        	    public void keyPressed( KeyEvent pEvent ) {
//        	    	if ( pEvent.getKeyCode() == KeyEvent.VK_ESCAPE ) {
//        	    		endAnyEditing( false );
//        	    	} else if ( pEvent.getKeyCode() == KeyEvent.VK_ENTER ) {
//        	    		endAnyEditing( true );
//        	    	} 
//        	    }
//        	}
//        );
//        
//        accessibleEditorStep = new JTextField();
//        
//        applyAccessibleEditorScheme( accessibleEditorStep );
//        
//        accessibleEditorStep.setCaretColor( INFO_COLOR );
//        
//        accessibleEditorIncrease = new JButton( "+" );
//        
//        applyAccessibleEditorScheme( accessibleEditorIncrease );
//        
//        accessibleEditorIncrease.setFocusPainted( false );
//        
//        accessibleEditorIncrease.addActionListener(
//        	new ActionListener() {
//				public void actionPerformed( ActionEvent pEvent ) {
//					
//				}
//        	}
//        );
//        
//        accessibleEditorDecrease = new JButton( "-" );
//
//        applyAccessibleEditorScheme( accessibleEditorDecrease );
//        
//        accessibleEditorDecrease.setFocusPainted( false );
//        
        add( accessibleEditor );
        
//        add( accessibleEditorStep );
//        add( accessibleEditorIncrease );
//        add( accessibleEditorDecrease );
    	
    }
    
//    private void applyAccessibleEditorScheme( JComponent pComponent ) {
//    	
//    	pComponent.setOpaque( true );
//    	pComponent.setBorder( BorderFactory.createLineBorder( INFO_COLOR ) );
//					
//    	pComponent.setBackground( Color.darkGray );
//    	pComponent.setForeground( INFO_COLOR );
//    	pComponent.setFont( INFO_FONT );
//    	
//    	pComponent.setVisible( false );
//    	
//    }

    private void initUpdateAndPaintingThreads() {
    	
    	painter = new Painter( getSize().width, getSize().height );
    	
    	Thread wPt = new Thread( painter );
    	
    	wPt.start();
    	
    	updater = new Updater();
    	
    	Thread wUt = new Thread( updater );
    	
    	wUt.setPriority( Thread.MAX_PRIORITY );
    	
    	wUt.start();
    	
    }
    
    public boolean isBusy() {
        return painter.isActive();
    }

    public boolean isSettingIterationsFromPalette() {
    	return setIterationsFromPalette;
    }
    
    public void paint( Graphics pGrp ) {

        // -- Verify if panel is ready to be painted.
        if ( !canRender() ) {
            return;
        }
        
        
        if ( ( getRenderer() != null && getRenderer().isModified() ) || forceRepaint ) {

        	painter.paint();

        }  else {

        	pGrp.drawImage( painter.getBuffer(), 0, 0, null );
            paintAccessibles( pGrp );
            
        }

//        super.paint( pGrp );
        
    }

    public void repaint() {
    	
        forceRepaint = true;
        
        super.repaint();
        
    }

    protected void resetAccessibleHandlers() {
    	
    	for( AccessibleHandler wAh : accessibleHandlers ) {
    		if ( wAh != null ) {
    			wAh.reset();
    		}
    	}
    	
    }
    
    public void setAccessibleContainer( FCAccessibleContainer pAccessibleContainer ) {
        accessibleContainer = pAccessibleContainer;
    }

    public FCFractal getFractal() {
        return (FCFractal)getAccessibleContainer().getAccessible(
            FCAccessibleContainer.ACCESSIBLE_FRACTAL );
    }

    public FCRenderer getRenderer() {
        return (FCRenderer)getAccessibleContainer().getAccessible(
            FCAccessibleContainer.ACCESSIBLE_RENDERER );
    }

    protected int getAbsoluteSize( boolean pSmallest ) {

        int w = getSize().width;
        int h = getSize().height;

        if ( pSmallest ) {

            if ( w < h ) {
                return w;
            } else {
                return h;
            }

        } else {

            if ( w > h ) {
                return w;
            } else {
                return h;
            }

        }

    }

    public FCAccessibleContainer getAccessibleContainer() {
        return accessibleContainer;
    }

    protected void setFractalOffset( int pX, int pY ) {

        double wXOff = getFractal().getXOffset();
        double wYOff = getFractal().getYOffset();

//        int wSz = Math.min( getWidth(), getHeight() );
        
        wXOff += getFractal().viewToSet( pX, getWidth() );
        wYOff += getFractal().viewToSet( pY, getHeight() );
        
        getFractal().setXOffset( wXOff );
        getFractal().setYOffset( wYOff );

    }

    public void setHideAccessibles( boolean pState ) {
    	hideAccessibles = pState;
    	super.repaint();
    }
    
    public void setIterationsFromPalette( boolean pState ) {
    	setIterationsFromPalette = pState;
    }
    
	public void keyPressed( KeyEvent pEvent ) {
		
//		System.out.println( KeyEvent.VK_ESCAPE + " " + pEvent );
		
		if ( pEvent.getKeyCode() == KeyEvent.VK_F5 ) {
			repaint();
		}
		
		if ( pEvent.getKeyCode() == KeyEvent.VK_ESCAPE ) {
			painter.stop();
		}
		
		// -- Show cross-hair if ALT it pressed.
		if ( pEvent.isAltDown() ) {
			if ( getCursor() == Cursor.getDefaultCursor() ) {
				setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
			}
		}
		
	}

	public void keyReleased( KeyEvent pEvent ) {

//		System.out.println( pEvent );
		
		if ( pEvent.getKeyCode() == KeyEvent.VK_ALT ) {
			if ( getCursor() == Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) ) {
				setCursor( Cursor.getDefaultCursor() );
			}
		}
		
	}

	public void keyTyped(KeyEvent e) {
		// TODO Auto-generated method stub
		
	}
    
    public void mouseClicked( MouseEvent pEvent ) {

//    	System.out.println( pEvent );
    	
        if ( !accessibleEvent && ( pEvent.getModifiers() & MouseEvent.BUTTON1_MASK ) > 0 ) {

            if ( pEvent.isControlDown() ) {
                
            	getFractal().setZoom( getFractal().getZoom() / 2.0 );
                
                repaint();
                
            } else {

                if ( pEvent.getClickCount() >= 2 ) {

                    int wSz = getAbsoluteSize( true ) >> 1;
                    
                    int wDx = pEvent.getPoint().x; // + ( wSz >> 1 );
                    int wDy = pEvent.getPoint().y; // + ( wSz >> 1 );
                   
//                    int wH = getSize().height;
//                    int wW = getSize().width;
//
//                    if ( wW > wH ) {
//                        wDx -= ( wW - wH ) << 1;
//                    } else if ( wH > wW ) {
//                        wDy -= ( wH - wW ) << 1;
//                    }
                    
                    setFractalOffset( wDx, wDy );

                    //double wDz = (double)getAbsoluteSize( true ) / (double)wSz;
                	getFractal().setZoom( getFractal().getZoom() * 2.0 );
                	
                	accessibleContainer.refresh();
                	
//                	painter.scroll(
//	                    wDx - ( getWidth() >> 1 ),
//	                    wDy - ( getHeight() >> 1 ) );
	                
                	
                    repaint();

                }

            }
            
        }

    }
    
    public void mousePressed( MouseEvent pEvent ) {

        if ( ( pEvent.getModifiers() & MouseEvent.BUTTON1_MASK ) > 0 ) {

        	if ( handleAccessibleEvent( pEvent ) ) {
        		accessibleEvent = true;
        	} else {
        		
        		accessibleEditor.stopEditing( true );
        		accessibleEvent = false;
        		
        		// -- Entry magnify mode.
        		if ( pEvent.isAltDown() ) {
        			magnify = true;
        		}
        		
        	}
        	
            dragX = pEvent.getPoint().x;
            dragY = pEvent.getPoint().y;

            ddx   = pEvent.getPoint().x;
            ddy   = pEvent.getPoint().y;

            fdox = getFractal().getXOffset();
            fdoy = getFractal().getYOffset();

        }

    }

    public void mouseReleased( MouseEvent pEvent ) {

        if ( magnify ) {

        	magnify = false;
        	
            Graphics wGrp = getGraphics();

            wGrp.setXORMode( Color.white );

            wGrp.drawRect(
                dragX,
                dragY,
                lastSz,
                lastSz );

            int wX = dragX + ( lastSz >> 1 );
            int wY = dragY + ( lastSz >> 1 );
            int wH = getSize().height;
            int wW = getSize().width;

//            if ( wW > wH ) {
//                wX += ( (double)( wW - wH ) / 2.0 );
//            } else if ( wH > wW ) {
//                wY += ( (double)( wH - wW ) / 2.0 );
//            }

            setFractalOffset( wX, wY );
  
            double wDz = (double)getAbsoluteSize( true ) / (double)lastSz;
            double wDs = getFractal().getZoom() * wDz;

            getFractal().setZoom( wDs );

            accessibleContainer.refresh();
            
//            forceRepaint = true;

            repaint();

        }

        lastSz = -1;

    }

    public void mouseEntered( MouseEvent pEvent ) { }

    public void mouseExited( MouseEvent pEvent ) { }

    public void mouseDragged( MouseEvent pEvent ) {

        if ( !accessibleEvent && ( pEvent.getModifiers() & MouseEvent.BUTTON1_MASK ) > 0 ) {

            if ( magnify ) {  

            	magnify = true;
            	
                int wDx = pEvent.getPoint().x - dragX;
                int wDy = pEvent.getPoint().y - dragY;

                int wSz = wDy;

                if ( wDx > wDy ) {
                    wSz = wDx;
                }

                Graphics wGrp = getGraphics();

                wGrp.setXORMode( Color.white );

                wGrp.drawRect( dragX, dragY, lastSz, lastSz );

                wGrp.drawRect( dragX, dragY, wSz, wSz );

                lastSz = wSz;

            } else {  // -- Offset drag

                int wXo = ( getWidth() >> 1 ) - ( pEvent.getX() - dragX );
                int wYo = ( getHeight() >> 1 ) - ( pEvent.getY() - dragY );

   //             System.out.println( wXo + " " + wYo );
                
                getFractal().setXOffset( fdox + getFractal().viewToSet( wXo, getWidth() ) );
                getFractal().setYOffset( fdoy + getFractal().viewToSet( wYo, getHeight() ) );

     //       	System.out.println( "( " + wXo + " - ( " + getWidth() + " / 2.0 ) ) /  " + getFractal().getZoom() + " = " + ( getFractal().viewToSet( wXo, getWidth() ) ) );
                
                ddx = ddx - pEvent.getX();
                ddy = ddy - pEvent.getY();

//                scrollBuffer( ddx, ddy );
                painter.scroll( ddx, ddy );

                ddx = pEvent.getX();
                ddy = pEvent.getY();

                accessibleContainer.refresh();
                
                repaint();

            }

        }


    }

    public void mouseMoved( MouseEvent pEvent ) { 
    	
    	handleAccessibleEvent( pEvent );
    	
//    	Graphics wGrp = getGraphics();
//
//    	if ( wGrp != null ) {
//    	
//	    	wGrp.fillRect( 20, 256 - 16, 256, 48 );
//	    	
//	    	wGrp.setColor( Color.GREEN );
//	    	
//	    	wGrp.drawString( 
//	    		String.valueOf( getFractal().viewToSet( pEvent.getX(), getWidth() ) + getFractal().getXOffset() ), 
//	    		20, 
//	    		256 );
//	    	
//	    	wGrp.drawString( 
//	    		String.valueOf( getFractal().viewToSet( pEvent.getY(), getHeight() ) + getFractal().getYOffset() ), 
//	    		20, 
//	    		272 );
//	    	
//    	}
    	
    }

    protected void processEvent( AWTEvent pEvent ) {
    	
       if ( !isBusy() || pEvent instanceof KeyEvent || pEvent.getID() == ComponentEvent.COMPONENT_RESIZED ) {
            super.processEvent( pEvent );
       }

    }

    public void componentResized( ComponentEvent pEvent ) {

    	resetAccessibleHandlers();
    	
    	painter.resize( pEvent.getComponent().getWidth(), pEvent.getComponent().getHeight() );
    	
//        repaint();

    }

    public void componentMoved( ComponentEvent pEvent ) { }

    public void componentShown( ComponentEvent pEvent ) { }

    public void componentHidden( ComponentEvent pEvent ) { }

    public RenderedImage createSnapshot( int pWidth, int pHeight ) {
    	
    	BufferedImage wBi = new BufferedImage( pWidth, pHeight, BufferedImage.TYPE_INT_ARGB );
    	
    	Graphics wGrp = wBi.createGraphics();
    	
        wGrp.setClip( new Rectangle( 0, 0, pWidth, pHeight ) );

        getRenderer().paint( wGrp );
    	
        return wBi;
        
    }
    
    public void update( Graphics pGraphics ) {
        //paint( pGraphics );
    }

    private class AccessibleObjectHandler {

    	public int x, y, w, dw, h, fa;
    	
    	public String  description;
    	public boolean highlighted;
    	public boolean right;
    	
    }
    
    private class AccessibleHandler implements FCAccessibleEditor.EditingListener {
    	
//    	private List accessibleObjectHandlers = new ArrayList();
    	
    	private FCAccessible accessible;
    	
    	private int position;

    	private AccessibleObjectHandler[] aohs;
    	
    	private FCAccessibleObject editing;
    	
    	public AccessibleHandler( FCAccessible pAccessible, int pPosition ) {
    		
    		accessible = pAccessible;
    		position = pPosition;
    		
    	}

    	public void editingCommitted( FCAccessibleEditor pEditor ) {
    		repaint();
    	}
    	
    	public void editingEnded( FCAccessibleEditor pEditor ) {
    	
    		if ( editing != null ) {
    		
    			accessibleEditor.setVisible( false );
    			accessibleEditor.setAccessibleObject( null );
    			
    			FCFractalPanel.this.requestFocus();
    			
    			editing = null;
				
    		}
    		
    	}
    	
        private void drawAccessibleText( Graphics pGrp, String pText, int pX, int pY, boolean pHi ) {

            pGrp.setColor( Color.black );
            pGrp.drawString( pText, pX + 1, pY + 1 );
            
            if ( pHi ) {
            	pGrp.setColor( INFO_HIGHLIGHT_COLOR );
            } else {
            	pGrp.setColor( INFO_COLOR );
            }
            
            pGrp.drawString( pText, pX, pY );

        }
        
        private void editAccessibleObject( final AccessibleObjectHandler pAoh, final FCAccessibleObject pAo ) {
        	
			if ( editing != pAo ) {
				accessibleEditor.stopEditing( true );
			}
			
			if ( pAo.getType() == boolean.class ) {
				
				pAo.setValue( !( (Boolean)pAo.getValue() ).booleanValue() );
				
				repaint();
				
				return;
				
			} 

			accessibleEditor.setEditingListener( this );
			
			if ( pAoh.right ) {

				accessibleEditor.setLocation( pAoh.x - 227, pAoh.y - 1 );
				
			} else {
				
				accessibleEditor.setLocation( pAoh.x + pAoh.dw - 1, pAoh.y - 1 );
				
			}

			accessibleEditor.setRightAligned( pAoh.right );
			accessibleEditor.setSize( 228, pAoh.h );
			
			accessibleEditor.setAccessibleObject( pAo );
			
			accessibleEditor.setVisible( true );
			
			accessibleEditor.requestFocus();
	        accessibleEditor.selectAll();
			
			editing = pAo;

        }
    	
    	public boolean handleEvent( MouseEvent pEvent ) {

    		if ( aohs == null ) {
    			return false;
    		}
    		
    		boolean sh = false;
    		boolean eh = false;
    		
        	Point p = pEvent.getPoint();
        	
            for ( int a = 0; a < aohs.length; a++ ) {

            	AccessibleObjectHandler wAoh = aohs[ a ];
            	
            	int x = ( position & POSITION_RIGHT ) != 0 ? wAoh.x - ( wAoh.w - wAoh.dw ) : wAoh.x;  
            	
            	if ( ( p.x >= x && p.x < x + wAoh.w ) && 
            			( p.y >= wAoh.y && p.y < wAoh.y + wAoh.h ) ) {
            		
            		if ( !wAoh.highlighted ) {
            			wAoh.highlighted = true;
            			sh = true;
            		}
            		
            		if ( pEvent.getButton() == MouseEvent.BUTTON1 && pEvent.getClickCount() == 1 ) {
            			editAccessibleObject( wAoh, ( (FCAccessibleObject)accessible.getAccessibleObjects().get( a ) ) );
            			eh = true;
            		}
            		
            	} else {
            		
            		if ( wAoh.highlighted ) {
            			wAoh.highlighted = false;
            			sh = true;
            		}
            		
            	}
            	
            }
    		
            if ( sh ) {
            	paintAccessibles( getGraphics() );
            }
			
//            if ( !eh && pEvent.getID() == MouseEvent.MOUSE_CLICKED ) {
//            	accessibleEditor.setVisible( false );
//            }
            
            return eh;
    		
    	}
    	
    	public synchronized boolean paint( Graphics pGrp ) {
    		
    		if ( aohs == null ) {
   				prepareAccessibleObjectHandlers( accessible, pGrp );
    		}
    		
            List<FCAccessibleObject> wAos = accessible.getAccessibleObjects();

            int x = 4, y = 18, w, dw = 0, h = y;
            
            boolean hl = false;
            boolean sh = false;

            // -- Draw title
            pGrp.setFont( INFO_FONT_BOLD );

            String wText = accessible.getDescription();
    	
            w = pGrp.getFontMetrics().getStringBounds( wText, pGrp ).getBounds().width;
            
            if ( ( position & POSITION_RIGHT ) != 0 ) {
                x = getWidth() - w - x;
            }

            if ( ( position & POSITION_BOTTOM ) != 0 ) {
                y = getHeight() - 6;
            }
            
            drawAccessibleText( pGrp, wText, x, y, false );
            
            // -- Draw accessibles.
            for ( int a = 0; a < wAos.size(); a++ ) {

                pGrp.setFont( INFO_FONT );

        		FCAccessibleObject wAo = (FCAccessibleObject)wAos.get( a );

        		if ( wAo != editing ) {
	        		wText = String.valueOf( wAo.getValue() );
            	} else {
            		wText = "";
            	}
	        		
                w = pGrp.getFontMetrics().getStringBounds( wText, pGrp ).getBounds().width;

                if ( ( position & POSITION_RIGHT ) != 0 ) {
                    x = aohs[ a ].x - w;
                    wText = wText + aohs[ a ].description;   
                } else { 
                    x = aohs[ a ].x;
                    wText = aohs[ a ].description + wText;   
                }
                
                aohs[ a ].w = aohs[ a ].dw + w;

//                System.out.println( aohs[ a ].description + " " + aohs[ a ].w );
                
                if ( aohs[ a ].highlighted ) {
                	sh = true;
                }
                
                drawAccessibleText( pGrp, wText, x, aohs[ a ].y + aohs[ a ].fa, aohs[ a ].highlighted );

            }
    		
        	return sh;
        	
    	}
    	
    	private void prepareAccessibleObjectHandlers( FCAccessible pAccessible, Graphics pGrp ) {
    		
            List wAos = pAccessible.getAccessibleObjects();
            
    		AccessibleObjectHandler[] wAohs = new AccessibleObjectHandler[ wAos.size() ];

            int x = 7, w = 0, dw = 0, h = 24;
            
    		Graphics wGrp = pGrp;
    		
			wGrp.setFont( INFO_FONT );
    		
			FontMetrics wFm = wGrp.getFontMetrics();
			
    		for( int a = 0; a < wAos.size(); a++ ) {
    			
    			AccessibleObjectHandler wAoh = new AccessibleObjectHandler();
    			
        		FCAccessibleObject wAo = (FCAccessibleObject)wAos.get( a );

        		String wDesc;
        		
                if ( ( position & POSITION_RIGHT ) != 0 ) {
                	wDesc = " :" + wAo.getDescription();
                	wAoh.right = true;
                } else {
                	wDesc = wAo.getDescription() + ": ";
                }
        		
           		dw = wFm.getStringBounds( wDesc, wGrp ).getBounds().width;

           		// -- Find the largest description width.
           		if ( dw > w ) {
           			w = dw;
           		}
           		
                if ( ( position & POSITION_BOTTOM ) != 0 ) {
                	wAoh.y = getHeight() - h;
                } else {
                	wAoh.y = h;
                }

                wAoh.description = wDesc;
                wAoh.dw = dw;
                wAoh.h  = 16;
                wAoh.fa = wFm.getAscent();

                h += 16;
                
                wAohs[ a ] = wAoh;
    			
    		}
    		
        	// -- Align 'X' values.
    		for( int a = 0; a < wAos.size(); a++ ) {
                
                if ( ( position & POSITION_RIGHT ) != 0 ) {
                	wAohs[ a ].x = getWidth() - ( ( x + wAohs[ a ].dw ) + ( w - wAohs[ a ].dw ) );
                } else {
                	wAohs[ a ].x = x + ( w - wAohs[ a ].dw );
                }
                
    		}
    		
    		aohs = wAohs;
        	
    	}
    	
    	public synchronized void reset() {
    		aohs = null;
    	}
    	
    }
    
    private class Painter implements Runnable {

    	private boolean active = false;
    	private boolean disposed = false;
    	private boolean mustPaint = false;
    	
    	private Image buffer;
    	
    	private int sdx, sdy;
    	private int h, w;
    	
        public Painter( int pWidth, int pHeight ) { 
 
        	w = pWidth;
        	h = pHeight;
        	
        }
        
        public synchronized Image getBuffer() {
        	return buffer;
        }

        public boolean isActive() {
        	return active;
        }
        
        protected synchronized void maintainImageBuffer() {
        	
        	if ( buffer == null || buffer.getWidth( null ) != w || buffer.getHeight( null ) != h ) {
        		
        		if ( buffer != null ) {
        			buffer.flush();
        		}
        		
        		buffer = createImage( w, h );
        		
//            	System.out.println( "IBS: " + w + " " + h );
            	
        	}
        
        	
        }
        
        public void run() {

        	int wSdx = 0, wSdy = 0;
        	boolean wMustScroll = false;

        	Rectangle wCr = new Rectangle();
        	
        	while( !disposed ) {
        	
	        	synchronized( Painter.this ) {
	        		
	        		active = false;
	        		
		            // -- Notify stop wait.
		            notify();
	        		
	        		// -- Wait until we are told to paint, unless 'mustPaint' is tripped.
		        	try {
		        //		if ( !mustPaint ) {
		        			Painter.this.wait();
		        //		}
		        	} catch( InterruptedException pEx ) {
		        		System.out.println( "Painter interrupted!" );
		        	}
        		
		        	active = true;
		       // 	mustPaint = false;
		        	
		        	maintainImageBuffer();
		        	
		        	// -- Prepare for scrolling.
		        	if ( sdx != 0 || sdy != 0 ) {
		        		
		        		wSdx = sdx;
		        		wSdy = sdy;
		        		
		        		sdx = 0;
		        		sdy = 0;
		        		
		        		wMustScroll = true;
		        		
		        	} else {
		        		wMustScroll = false;
		        	}
		        	
		        	// -- Apply all accessible values.
		        	accessibleContainer.apply();
		        	
	        	}

	        	// -- Start updater.
	        	updater.start();
	        	
	            Graphics wGrp = getBuffer().getGraphics();
	            
	            FCRenderer wRenderer = getRenderer();
	
	            try {
	
//	                wRenderer.setViewportSize( getAbsoluteSize( false ) );
	                wRenderer.setViewportSize( getSize() );
	                
//	                wRenderer.getFractal().applyAccessible();
//	                wRenderer.applyAccessible();
	
	                if ( wMustScroll ) {
	
//	                	System.out.println( "SB: " + wSdx + " " + wSdy );

	                    int wAx = Math.abs( wSdx );
	                    int wAy = Math.abs( wSdy );
	
	                    int wScx = ( wSdx >= 0 ) ? wAx : 0;
	                    int wScy = ( wSdy >= 0 ) ? wAy : 0;
	
	                    wGrp.copyArea(
	                        wScx,
	                        wScy,
	                        getBuffer().getWidth( null ) - wAx,
	                        getBuffer().getHeight( null ) - wAy,
	                        -wSdx,
	                        -wSdy );
	
	                    if ( wSdx > 0 ) {
	                        wCr.x = getBuffer().getWidth( null ) - wAx;
	                    } else {
	                        wCr.x = 0;
	                    }
	
	                    wCr.y = 0; //( wSdy > 0 ) ? 0 : wAy;
	                    wCr.height = getBuffer().getHeight( null );
	                    wCr.width = wAx;
	
	                    wGrp.setClip( wCr );

	                    if ( wCr.width > 0 ) {
	                    	wRenderer.paint( wGrp );
	                    }
	
	                    if ( wSdy > 0 ) {
	                        wCr.y = getBuffer().getHeight( null ) - wAy;
	                    } else {
	                        wCr.y = 0;
	                    }
	
	                    wCr.x = 0; //( wSdx > 0 ) ? 0 : wAx;
	                    wCr.height = wAy;
	                    wCr.width = getBuffer().getWidth( null ) - ( ( wSdx > 0 ) ? wAx : 0 );
	
	                    wGrp.setClip( wCr );

	                    if ( wCr.height > 0 ) {
	                    	wRenderer.paint( wGrp );
	                    }
	
	                    wMustScroll = false;
	                    
	                } else {
	
	                    wGrp.setClip( new Rectangle(
	                        0,
	                        0,
	                        getBuffer().getWidth( null ),
	                        getBuffer().getHeight( null ) ) );
	
	                    wRenderer.paint( wGrp );
	
	                }
	
	            } catch( Exception pEx ) {
	
	
	                // -- Ignore rendering errors.
	            	pEx.printStackTrace();
	

	            }

	            synchronized( Painter.this ) {

	     //       	System.out.println( "CM!" );
	                wRenderer.clearModified();
	                
		            forceRepaint = false;
		            updater.stop();
		            
	            }
	            
        	}
        	
        }
        
        public synchronized void paint() {

    //		mustPaint = true;
    		
        	if ( !isActive() ) {
        		Painter.this.notify();
        	}
        	
        }

        public synchronized void resize( int pWidth, int pHeight ) {
        	
        	stop( true );
        	
        	w = pWidth;
        	h = pHeight;

        	paint();
        	
        }
        
        public synchronized void scroll( int pSdx, int pSdy ) {

        	sdx += pSdx;
        	sdy += pSdy;
        	
        }
        
        public void stop() {
        	stop( false );
        }
        
        public synchronized void stop( boolean pWait ) {
        	
        	getRenderer().stop();
        	
        	if ( pWait && isActive() ) {
        		try {
        			wait();
        		} catch( InterruptedException pEx ) {
        			// -- Do nothing.
        		}
        	}
        	
        }
        
    }

    private class Updater implements Runnable {

    	private boolean disposed = false;
    	private boolean updating = false;
    	
        public synchronized boolean isActive() {
        	return updating;
        }
        
        public void run() {

        	while( !disposed ) {
        	
        		synchronized( Updater.this ) {
        		
	        		try {
	        			Updater.this.wait();
	        		} catch( InterruptedException pEx ) {
	        			System.out.println( "Updater interrupted!" );
	        		}

	        		updating = true;
	        		
		            try {
		            	
			            while( updating ) {
			
	//		            	System.out.println( "U: " + updating + " PA: " + painter.isActive() );
			            	
			                // -- Painter will interrupt this thread when it's done.
			                // -- Prevent image from being refreshed twice.
		            		Updater.this.wait( 1000 );
		            		
		                	if ( updating ) {
		
		               // 		System.out.println( "UD!" );
		                		
			            		FCFractalPanel.this.setCursor( Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ) );
			
			                    FCFractalPanel.this.getGraphics().drawImage( painter.getBuffer(), 0, 0, null );
			                    paintAccessibles( FCFractalPanel.this.getGraphics() );
			    	        	
		                	}
				
			            }
		
		                FCFractalPanel.this.getGraphics().drawImage( painter.getBuffer(), 0, 0, null );
		                paintAccessibles( FCFractalPanel.this.getGraphics() );
		                
		            } catch( Exception pEx ) { 
		            	
		            	pEx.printStackTrace();
		            	
		            }
		        		            	
		            FCFractalPanel.this.setCursor( Cursor.getDefaultCursor() );
		            
		            updating = false;
	            
        		}
        		
        	}

        }
        
        public synchronized void start() {

        	if ( !updating ) {
        		Updater.this.notify();
        	}
        	
        }
        
        public synchronized void stop() {
        	
   //     	System.out.println( "US2" );
        	
        	updating = false;

        	Updater.this.notify();

        }

    }


}