/* ----------------------------------------------------------------------------
 * File:       fc/fractal/FCJuliaFractal.java
 *
 * Author:     Unknown
 *
 * Copyright (c) - 1999, Technologies Interactives Mediagrif Inc.
 *
 * History
 * ----------------------------------------------------------------------------
 *  Date         User               Description
 * ----------------------------------------------------------------------------
 * 2000/02/11   Kris Fudalewski     Creation.
 * ----------------------------------------------------------------------------
 */
package fc.fractal;

import java.math.BigDecimal;
import java.math.MathContext;
import java.util.Properties;

public class FCMandelbrotFractal extends FCFractal {

	// -- Precision setting for HP mandelbrot.
	private static final MathContext mc = MathContext.DECIMAL128;
	
    private short wIter = 0;

    private double wImag = 0; 
    private double wReal = 0; 
    private double wImagSq = 0;
    private double wRealSq = 0;
    
    private double wX = 0; 
    private double wY = 0; 
    
    private boolean highPrecision = false;
    private double  multiplier;
    private double  velocityLimit;
    
    private BigDecimal multiplierHp;
    private BigDecimal velocityLimitHp;
    private BigDecimal zoomHp;
	
    public FCMandelbrotFractal( Properties pProp ) {
        super( pProp );
    }

    public short computePoint( double pX, double pY ) {

        wIter = 0;
        
        wImag = 0; 
        wReal = 0; 
        wImagSq = 0;
        wRealSq = 0;
    	
    	wX = xOffset + pX;    	
    	wY = yOffset + pY;
    	
        do {
        	
            wRealSq = wReal * wReal; 
            wImagSq = wImag * wImag;
        	
            wImag = ( ( wReal * wImag * multiplier ) ) + wY;
            wReal = ( wRealSq - wImagSq ) + wX;

            wIter++;

        } while ( ( wRealSq + wImagSq < velocityLimit ) && wIter < iterations );

// Mandelbrot ^ 3
//        do {
//        	
//            wRealSq = Math.pow( wImag, 3.0 ); 
//            wImagSq = wImag * wImag;
//        	
//            wImag = ( ( 3.0 * Math.pow( wReal, 2.0 ) * wImag ) - Math.pow( wImag, 3.0 ) ) + wY;
//            wReal = ( wRealSq - ( 3.0 * wReal * wImagSq ) ) + wX;
//
//            wIter++;
//
//        } while ( ( wRealSq + wImagSq < velocityLimit ) && wIter < iterations );
        
        return wIter;
        
    }
    
    public short computeViewPoint( int pX, int pY, int pXVSz, int pYVSz ) {
    	
    	if ( highPrecision ) {
    		return computePointHP( viewToSetHP( pX, pXVSz ), viewToSetHP( pY, pYVSz ) );
    	} else {
    		return super.computeViewPoint( pX, pY, pXVSz, pYVSz );
    	}
    	
    }
    
    protected short computePointHP( BigDecimal pX, BigDecimal pY ) {

//        double wImag = 0; //pX + xOffset;
//        double wReal = 0; //pY + yOffset;
//        double wImagSq = 0;
//        double wRealSq = 0;

    	wIter = 0;
    	
        BigDecimal wImag   = new BigDecimal( 0.0, mc );
        BigDecimal wReal   = new BigDecimal( 0.0, mc );
        BigDecimal wImagSq = new BigDecimal( 0.0, mc );
        BigDecimal wRealSq = new BigDecimal( 0.0, mc );
        
        BigDecimal wX = pX.add( new BigDecimal( xOffset, mc ), mc ); 
        BigDecimal wY = pY.add( new BigDecimal( yOffset, mc ), mc );

        do {

            wRealSq = wReal.multiply( wReal, mc ); 
            wImagSq = wImag.multiply( wImag, mc );  
            
            wImag = wImag.multiply( wReal.multiply( multiplierHp, mc ), mc ).add( wY, mc ); // ( 2.0 * wReal * wImag ) + wY;
            wReal = wRealSq.subtract( wImagSq, mc ).add( wX, mc );

            wIter++;

        } while ( ( ( wRealSq.add( wImagSq, mc ) ).compareTo( velocityLimitHp ) ) < 0 && wIter < iterations );

        return wIter;

    }

    public double getMultiplier() {
    	return multiplier;
    }
    
    public double getVelocityLimit() {
    	return velocityLimit;
    }
    
    public boolean isHighPrecision() {
    	return highPrecision;
    }
    
    public void setHighPrecision( boolean pState ) {
    	highPrecision = pState;
    }
    
    public void setMultiplier( double pMultiplier ) {
    	multiplier = pMultiplier;
    	multiplierHp = new BigDecimal( pMultiplier, mc );
    }
    
    public void setVelocityLimit( double pVelocityLimit ) {
    	velocityLimit = pVelocityLimit;
    	velocityLimitHp = new BigDecimal( pVelocityLimit, mc );
    }
    
    public void setZoom( double pZoom ) {
    	
    	super.setZoom( pZoom );
    	
    	zoomHp = new BigDecimal( pZoom, mc );
    	
    }
    
    protected BigDecimal viewToSetHP( int pValue, int pSize ) {
        return new BigDecimal( pValue - ( pSize >> 1 ), mc ).divide( zoomHp, mc );
    }

}

