/* ----------------------------------------------------------------------------
 * File:       fc/renderer/FCHeightFieldRenderer.java
 *
 * Author:     Unknown
 *
 * Copyright (c) - 1999, Technologies Interactives Mediagrif Inc.
 *
 * History
 * ----------------------------------------------------------------------------
 *  Date         User               Description
 * ----------------------------------------------------------------------------
 * 2000/02/18   Kris Fudalewski     Creation.
 * ----------------------------------------------------------------------------
 */
package fc.renderer;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Properties;

public class FCHeightFieldRenderer extends FCRenderer {

    protected short[][] map;

    public double  heightRatio = 0.25;
    private boolean paletteToRatio = true;
    private boolean inward;

    public FCHeightFieldRenderer( Properties pProp ) {
        super( pProp );
    }

    private void createMap( int pWidth, int pHeight ) {
        map = new short[ pHeight ][ pWidth ];
    }

    public double getHeightRatio() {
    	return heightRatio;
    }
    
    public boolean getPaletteToRatio() {
    	return paletteToRatio;
    }
    
    public String getName() {
        return "Height Field";
    }

    public void drawPoint( Graphics pGrp, int pX, int pY, int pItr ) {

        int i = getFractal().getIterations();
        int h = pGrp.getClipBounds().height;

        int c = 0;
        int wCount = 0;
        int wDl = getDetailLevel();

        pItr *= heightRatio;
        
        if ( inward ) {

            for( int a = pItr - ( wDl - 1 ); a <= i && pY + a < h; a++ ) {

                int r = pY + a;

                if ( wCount <= wDl ) {
                    c = pItr;
                } else {
                    c = a;
                }

                drawLine( pGrp, pX, r, c, pItr, wDl );

                wCount++;

            }

        } else {

            for( int a = -pItr - ( wDl - 1 ); a <= 0 && pY + a < h; a++ ) {

                int r = pY + a;

                if ( wCount <= wDl ) {
                    c = -pItr + i;
                } else {
                    c = a + i;
                }

                drawLine( pGrp, pX, r, c, pItr, wDl );

                wCount++;

            }

        }

    }

    private void drawLine( Graphics pG, int pX, int pY, int pC, int pItr, int pDl ) {

        if ( pY >= 0 && pY < map.length && map[ pY ][ pX ] == 0 ) {

        	if ( paletteToRatio ) {
        		pG.setColor( getPalette().getColor( (int)( pC / heightRatio ) ) );
        	} else {
        		pG.setColor( getPalette().getColor( pC ) );
        	}
            
            pG.drawLine( pX, pY, pX + pDl - 1, pY );

            map[ pY ][ pX ] = (short)pItr;

        }

    }

    public void paint( Graphics pGrp ) {

        super.paint( pGrp );

        // -- Get clipping bounds
        Rectangle wBounds = pGrp.getClipBounds();

        createMap( wBounds.width, wBounds.height );

        int wDetailLevel = getDetailLevel();
        int wItr = getFractal().getIterations();
        Dimension wVSz = getViewportSize();

        int iw = inward ? 0 : wItr;

        stop = false;
        
        // -- Render details
        for ( int y = wBounds.height + iw - 1; y >= -wItr + iw; y -= wDetailLevel ) {

            for( int x = 0; x < wBounds.width; x += wDetailLevel ) {

                int wX = wBounds.x + x;
                int wY = wBounds.y + y;

                short i = getFractal().computeViewPoint( wX, wY, wVSz.width, wVSz.height );

                drawPoint( pGrp, x, y, i );

            }

            if ( stop ) {
            	break;
            } else {
            	Thread.yield();
            }

        }

    }

    public boolean isRenderingInward() {
        return inward;
    }

    public void setHeightRatio( double pHeightRatio ) {
    	heightRatio = pHeightRatio;
    }
    
    public void setPaletteToRatio( boolean pState ) {
    	paletteToRatio = pState;
    }
    
    public void setRenderingInward( boolean pState ) {
        inward = pState;
    }

}
