package fc.renderer;

import java.awt.Color;
import java.util.Properties;
import java.util.Vector;

import fc.accessible.FCAccessible;
import fc.accessible.FCAccessibleContainer;

public class FCPalette extends FCAccessible {

    private static final String PROP_PALETTE_CYCLE      = "palette.cycle";
    private static final String PROP_PALETTE_DESC       = "palette.description";
    private static final String PROP_PALETTE_SEQ_PREFIX = "palette.sequence.";

    private Color[] colors;

    public FCPalette( Properties pProps ) {

        super( pProps );

        parsePaletteFile( pProps );

    }

    private Color createColorSequence( Properties pProps, String pName,
            Vector pColors, Color pLast ) {

        int   wLen;
        Color wStart = pLast;
        Color wEnd;

        try {

            wLen = Integer.parseInt( pProps.getProperty( pName + ".length" ) );

            if ( wLen <= 0 ) {
                wLen = 1;
            }

        } catch( Exception pEx ) {
            wLen = 32;
        }

        try {
            wStart = Color.decode( pProps.getProperty( pName + ".start" ) );
        } catch ( Exception pEx ) {
            wStart = pLast;
        }

        try {
            wEnd = Color.decode( pProps.getProperty( pName + ".end" ) );
        } catch ( Exception pEx ) {
            wEnd = null;
        }

        if ( wStart == null ) {
            throw new IllegalArgumentException(
                "Invalid or missing sequence START color: " + pName );
        }

        if ( wEnd == null ) {
            throw new IllegalArgumentException(
                "Invalid or missing sequence END color: " + pName );
        }

        double wSr = (double)wStart.getRed();
        double wSg = (double)wStart.getGreen();
        double wSb = (double)wStart.getBlue();

        double wEr = (double)wEnd.getRed();
        double wEg = (double)wEnd.getGreen();
        double wEb = (double)wEnd.getBlue();

        double wFr = ( wEr - wSr ) / (double)wLen;
        double wFg = ( wEg - wSg ) / (double)wLen;
        double wFb = ( wEb - wSb ) / (double)wLen;

    //    System.out.println( wFr );
    //    System.out.println( wFg );
    //    System.out.println( wFb );

        for ( int a = 0; a < wLen; a++ ) {

            pColors.addElement( new Color( (int)wSr, (int)wSg, (int)wSb ) );

            wSr += wFr;
            wSg += wFg;
            wSb += wFb;

        }

        return wEnd;

    }

    protected String parseDescription( Properties pProp ) {
       	return pProp.getProperty( PROP_PALETTE_DESC );
    }
    
    private void parsePaletteFile( Properties pProps ) {

        Vector wColors = new Vector();

        Color  wLast  = null;
        int    wCount = 0;
        String wProp;

        while ( ( wProp = pProps.getProperty( PROP_PALETTE_SEQ_PREFIX + wCount ) ) != null ) {

            wLast = createColorSequence( pProps, wProp, wColors, wLast );
            wCount++;

        }

        boolean wCycle = Boolean.valueOf( pProps.getProperty( PROP_PALETTE_CYCLE ) ).booleanValue();

        // -- Cycle back palette
        if ( wCycle ) {

            int wSize = wColors.size();

            for ( int a = 1; a < wSize; a++ ) {
                wColors.add( wColors.get( wSize - a ) );
            }

        }

        colors = new Color[ wColors.size() ];
        wColors.copyInto( colors );

    }

    public Color getColor( int pIdx ) {
        return colors[ pIdx % colors.length ];
    }

    public Color[] getColors() {
        return colors;
    }

    public int getLength() {
        return colors.length;
    }

    public int getType() {
    	return FCAccessibleContainer.ACCESSIBLE_PALETTE;
    }
    
}