/* ----------------------------------------------------------------------------
 * File:       fc/renderer/FCRenderer.java
 *
 * Author:     Unknown
 *
 * Copyright (c) - 1999, Technologies Interactives Mediagrif Inc.
 *
 * History
 * ----------------------------------------------------------------------------
 *  Date         User               Description
 * ----------------------------------------------------------------------------
 * 2000/02/18   Kris Fudalewski     Creation.
 * ----------------------------------------------------------------------------
 */
package fc.renderer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Properties;

import fc.accessible.FCAccessible;
import fc.accessible.FCAccessibleContainer;
import fc.fractal.FCFractal;

public abstract class FCRenderer extends FCAccessible {

    protected FCFractal fractal;
    protected FCPalette palette;

    private Dimension viewportSize;
    
    public int detailLevel;

    protected boolean stop = false;
    
    public FCRenderer( Properties pProp ) {
        super( pProp );
    }

    public void paint( Graphics pGrp ) {

//    	fractal.applyAccessible();
    	
        if ( clearBeforeRender() ) {
            Rectangle wRect = pGrp.getClipBounds();
            pGrp.setColor( Color.black );
            pGrp.fillRect( 0, 0, wRect.width, wRect.height );
        }

    }

    public boolean clearBeforeRender() {
        return true;
    }

    public void drawPoint( Graphics pGrp, int pX, int pY, int pItr ) {

        pGrp.setColor( getPalette().getColor( pItr ) );

        int wDl = getDetailLevel();

        if ( wDl == 1 ) {
            pGrp.drawLine( pX, pY, pX, pY );
        } else {
            pGrp.fillRect( pX, pY, wDl, wDl );
        }

    }

    public void drawBlock( Graphics pGrp, int pX, int pY, int pSz, int pItr ) {

        pGrp.setColor( getPalette().getColor( pItr ) );
        pGrp.fillRect( pX, pY, pSz, pSz );

    }

    public void clearModified() {
        super.clearModified();
        getFractal().clearModified();
    }

    public boolean isModified() {

        if ( getFractal() != null && getFractal().isModified() ) {
            return true;
        }

        return super.isModified();

    }

    public Dimension getViewportSize() {
        return viewportSize;
    }

    public void setViewportSize( Dimension pViewportSize ) {
        viewportSize = pViewportSize;
    }

    public int getDetailLevel() {
        return detailLevel;
    }

    public void setDetailLevel( int pDetailLevel ) {
        detailLevel = pDetailLevel;
    }

    public FCFractal getFractal() {
        return (FCFractal)getAccessibleContainer().getAccessible(
            FCAccessibleContainer.ACCESSIBLE_FRACTAL );
    }

    public FCPalette getPalette() {
        return (FCPalette)getAccessibleContainer().getAccessible(
            FCAccessibleContainer.ACCESSIBLE_PALETTE );
    }
    
    public int getType() {
    	return FCAccessibleContainer.ACCESSIBLE_RENDERER;
    }

    public void stop() {
    	stop = true;
    }
    
//    public static interface Listener {
//    	
//    	void started( int pTotalLines );
//    	void rendered( int pLine, int pTotalLines );
//    	void ended( int pLine, int pTotalLines );
//    	
//    }
    
}